# -*- coding: utf-8 -*-
"""
Created on Wed Mar 25 10:51:20 2015

@author: adrienkuntz
"""

import numpy as np
from scipy import integrate
from scipy.special import sici
from scipy.optimize import brentq
#import matplotlib.pyplot as plt
import Trispectrum_PT as pt
import kernels as knl
from Params import *
import math




ktab = np.logspace(kmin_log, kmax_log, knum)



#function to obtain multidimensional arrays

def mafonction(ufct, *vectors):
    vs = np.ix_(*vectors)
    r = ufct.identity
    for v in vs:
        r = ufct(r,v)
    return r
    




###Powerspectrum
    
    
def WTH(x): return(3 / x**3 * (np.sin(x) - x * np.cos(x)))
def Wprime(x): return 3 / x**4 * (np.sin(x) * (x**2-3) + 3 * x * np.cos(x))    #Fourier transform of a top hat and its derivative

def Pasympt(k) : return math.pow(k, nspectral-4) * np.log(k)**2
    
def omegam(z) : return omegam0 * (1+z)**3 / knl.E(z)**2
def rho(z) : return rhobar * omegam(z) / omegam0                #comoving density as a function of z
def delta(z) : return (18*np.pi**2 + 82*(omegam(z)-1) - 39*(omegam(z)-1)**2) / omegam(z)        #delta of spherical collapse
    
for z in ztab :
    
    powerspectrum = pt.powerspectrum.copy()
    
    print 'z =', z
    
    powerspectrum[:, 1] = np.array(map(lambda k: pt.P(k, z), pt.powerspectrum[:,0]))

    
    def Rf(m) : return np.exp(1/3. * np.log(3. * m / (4. * np.pi * rho(z))))        #Radius within which the mass is m
        
    ###Definition of sigma2. To have a good value, the integration must be up to k*R(m)~10 because k**2*P(k) is very slowly convergent
        
    kmax_sigma = xmax / Rf(mmin)
    if kmax_sigma > powerspectrum[-1, 0] :             ###There is not enough points, complete P(k) with an asymptotic value in log space
    
        step = np.log10(powerspectrum[1, 0]) - np.log10(powerspectrum[0, 0])
        k0 = powerspectrum[-1, 0]
        Pk0 = powerspectrum[-1, 1]
        k = k0 * math.pow(10, step)
        
        while k <= kmax_sigma :
            ligne = np.array([[k, Pasympt(k) * Pk0 / Pasympt(k0)]])
            powerspectrum = np.concatenate((powerspectrum, ligne))
            k = k * math.pow(10, step)
         
         
    pow_matrix = mafonction(np.multiply, np.ones(mnum), powerspectrum[:,1])
    k_matrix = mafonction(np.multiply, np.ones(mnum), powerspectrum[:,0])
    
    
    def sigma2f(m):         #variance of the initial density field
        f = powerspectrum[:,0]**2 / (2 * np.pi**2) * powerspectrum[:,1] * WTH(powerspectrum[:,0] * Rf(m))**2
        return integrate.simps(f, x=powerspectrum[:,0])
        
    
    def nuf(m): return (delta_sc)**2 / sigma2f(m)
    
    
    mstar = brentq(lambda m:nuf(m)-1., mmin, 1e14)       #critical mass : nu(mstar)=1
       
    m_min_log = np.log10(mmin)
    m_max_log = np.log10(mstar) + 6
    mtab = np.logspace(m_min_log, m_max_log, mnum)
   
    
    
    ###Coefficients b (link the overdensity of halos to the overdensity of dark matter)
        

    
    R = np.exp(1/3. * np.log(3. * mtab / (4. * np.pi * rho(z))))
    kR = mafonction(np.multiply, R, powerspectrum[:,0])                               #k*R
    dRdm = 1./3 * np.exp(1/3. * np.log(3. / (4. * np.pi * rho(z) * mtab**2)))            #dR/dm
    
    
    
    
    
    
    sigma2 = integrate.simps(k_matrix**2 / (2 * np.pi**2) * pow_matrix * WTH(kR)**2, x=powerspectrum[:,0], axis=1)
        
    nu = (delta_sc)**2/sigma2   
    
        
        
    epsilon1 = (q * nu - 1) / (delta_sc)
    epsilon2 = q * nu * (q * nu - 3) / (delta_sc)**2
    epsilon3 = q * nu * (q * nu - 3)**2 / (delta_sc)**3
    
    E1 = 2 * p / (delta_sc) / (1 + np.exp(p * np.log(q * nu)))
    E2 = E1 * (2 * epsilon1 + (1 + 2 * p) / (delta_sc))
    E3 = E1 * ((4*(p**2-1) + 6*p*q*nu) / delta_sc**2 + 3*epsilon1**2)
        
    b1 = 1. + epsilon1 + E1
    b2 = 2 * (1+a2) * (epsilon1 + E1) + epsilon2 + E2
    b3 = 6 * (a2+a3) * (epsilon1 + E1) + 3 * (1+2*a2) * (epsilon2 + E2) + epsilon3 + E3
       
       
    
    
    
    
    ###Sheth-Tormen function with a good normalization factor (adapted to the range of integration considered here)
    
    Sheth = 0.3222 * (1 + np.exp(-p * np.log(q * nu))) * np.sqrt(q / (2 * np.pi * nu)) * np.exp(-q * nu / 2.)
    
    normfactor1 = integrate.simps(Sheth, x=nu)
    Sheth = Sheth / normfactor1
    
    normfactor2 = integrate.simps(Sheth * b1, x=nu)
    b1 = b1 / normfactor2
    
    normfactor3 = integrate.simps(Sheth * b2, x = nu)
    b2 = b2 - normfactor3
    
    normfactor4 = integrate.simps(Sheth * b3, x = nu)
    b3 = b3 - normfactor4
    
    ###density of halos
    
    
    dsigma2dm = dRdm * integrate.simps(k_matrix**3 / (np.pi**2) * pow_matrix * WTH(kR) * Wprime(kR), x=powerspectrum[:,0], axis=1) 
        
    
    numberdens = - rho(z) / mtab * Sheth * dsigma2dm * (delta_sc)**2 / sigma2**2   #convert to Mpc
        
        
        
        
    ###profile of a halo : NFW profile
        
        
    cbar = 9. / (1+z) * np.exp(-0.13 * np.log(mtab/mstar))    #concentration parameter
    rs = R / math.pow(delta(z), 1/3.) / cbar                            #radius parameter
    
    
    def profile(k):
        
        (si1, ci1) = sici((1+cbar) * k * rs)
        (si2, ci2) = sici(k * rs)
        u0 = 1/(np.log(1 + cbar) - cbar/(1+cbar))
        return u0 * (np.sin(k * rs) * (si1 - si2) - np.sin(k * rs * cbar)/((1 + cbar) * k * rs) + np.cos(k * rs) * (ci1 - ci2))
            
    
    
    
    ###Functions Mijn
    

    
    MF04 = lambda k1, k2, k3, k4 : integrate.simps(numberdens * (mtab / rho(z))**4 * profile(k1) * profile(k2) * profile(k3) * profile(k4), x = mtab)
    
    MF02 = lambda k1, k2 : integrate.simps(numberdens * (mtab / rho(z))**2 * profile(k1) * profile(k2), x = mtab)    
    
       
    
    MF11 = lambda k1 : integrate.simps(numberdens * (mtab / rho(z)) * b1 * profile(k1), x = mtab)
        
    MF12 = lambda k1, k2 : integrate.simps(numberdens * (mtab / rho(z))**2 * b1 * profile(k1) * profile(k2), x = mtab)
    
    MF13 = lambda k1, k2, k3 : integrate.simps(numberdens * (mtab / rho(z))**3 * b1 * profile(k1) * profile(k2) * profile(k3), x = mtab)
        
    MF14 = lambda k1, k2, k3, k4 : integrate.simps(numberdens * (mtab / rho(z))**4 * b1 * profile(k1) * profile(k2) * profile(k3) * profile(k4), x = mtab)
    
    
        
    MF21 = lambda k1 : integrate.simps(numberdens * (mtab / rho(z)) * b2 * profile(k1), x = mtab)
        
    MF22 = lambda k1, k2 : integrate.simps(numberdens * (mtab / rho(z))**2 * b2 * profile(k1) * profile(k2), x = mtab)
    
    MF23 = lambda k1, k2, k3 : integrate.simps(numberdens * (mtab / rho(z))**3 * b2 * profile(k1) * profile(k2) * profile(k3), x = mtab)
    
    MF24 = lambda k1, k2, k3, k4 : integrate.simps(numberdens * (mtab / rho(z))**4 * b2 * profile(k1) * profile(k2) * profile(k3) * profile(k4), x = mtab)



    MF31 = lambda k1 : integrate.simps(numberdens * (mtab / rho(z)) * b3 * profile(k1), x = mtab)
        


    
    ###Storage of the values of M in arrays. Only the particular points needed for the trispectrum are stored.
    
    
 
    
    MT11_k1 = np.array(map(MF11, ktab))
        
    MT21_k1 = np.array(map(MF21, ktab))
    
    MT31_k1 = np.array(map(MF31, ktab))

    
    
    MT12_k1k2 = np.zeros((knum, knum))
    for i in range(knum):
        for j in range(knum):
            MT12_k1k2[i,j] = MF12(ktab[i], ktab[j])

            
    MT22_k1k2 = np.zeros((knum, knum))
    for i in range(knum):
        for j in range(knum):
            MT22_k1k2[i,j] = MF22(ktab[i], ktab[j])
    
    
    MT04_k1k1k2k2 = np.zeros((knum, knum))
    for i in range(knum):
        for j in range(knum):
            MT04_k1k1k2k2[i,j] = MF04(ktab[i], ktab[j], ktab[i], ktab[j])
    
    
    MT13_k1k1k2 = np.zeros((knum, knum))
    for i in range(knum):
        for j in range(knum):
            MT13_k1k1k2[i,j] = MF13(ktab[i], ktab[i], ktab[j])
    
    
    MT02_k1k1 = np.zeros(knum)
    for i in range(knum):
        MT02_k1k1[i] = MF02(ktab[i], ktab[i])

    
    
    
    ###Write the result in files
    
    
    M11_k1 = open('M11_k1_{}.dat'.format(z), 'w')
    for i in range(knum) :
        M11_k1.write('{} {}\n'.format(ktab[i], MT11_k1[i]))
    M11_k1.close()
        
    M21_k1 = open('M21_k1_{}.dat'.format(z), 'w')
    for i in range(knum) :
        M21_k1.write('{} {}\n'.format(ktab[i], MT21_k1[i]))
    M21_k1.close()
    
    M31_k1 = open('M31_k1_{}.dat'.format(z), 'w')
    for i in range(knum) :
        M31_k1.write('{} {}\n'.format(ktab[i], MT31_k1[i]))
    M31_k1.close()
            
    M02_k1k1 = open('M02_k1k1_{}.dat'.format(z), 'w')
    for i in range(knum) :
        M02_k1k1.write('{} {}\n'.format(ktab[i], MT02_k1k1[i]))
    M02_k1k1.close()    
    
    
    
    M12_k1k2 = open('M12_k1k2_{}.dat'.format(z), 'w')
    for i in range(knum) :
        for j in range(knum) :
            M12_k1k2.write('{} {} {}\n'.format(ktab[i], ktab[j], MT12_k1k2[i,j]))
    M12_k1k2.close()
        
    M22_k1k2 = open('M22_k1k2_{}.dat'.format(z), 'w')
    for i in range(knum) :
        for j in range(knum) :
            M22_k1k2.write('{} {} {}\n'.format(ktab[i], ktab[j], MT22_k1k2[i,j]))
    M22_k1k2.close()
        
    M04_k1k1k2k2 = open('M04_k1k1k2k2_{}.dat'.format(z), 'w')
    for i in range(knum) :
        for j in range(knum) :
            M04_k1k1k2k2.write('{} {} {}\n'.format(ktab[i], ktab[j], MT04_k1k1k2k2[i,j]))
    M04_k1k1k2k2.close()
        
    M13_k1k1k2 = open('M13_k1k1k2_{}.dat'.format(z), 'w')
    for i in range(knum) :
        for j in range(knum) :
            M13_k1k1k2.write('{} {} {}\n'.format(ktab[i], ktab[j], MT13_k1k1k2[i,j]))
    M13_k1k1k2.close()    


# -*- coding: utf-8 -*-
"""
Created on Tue Mar 24 12:30:48 2015

@author: adrienkuntz
"""

##### Calculation of the bispectrum and the trispectrum in perturbation theory
##### The trispectrum is evaluated in points (k1, -k1, k2, -k2) for the purpose of the study
##### In this program uij means cos(theta(i,j)) with theta(i,j) the angle between ki and kj

import numpy as np
#from scipy import integrate
#import matplotlib.pyplot as plt
#import math
from Params import *


### Physical parameters

#omegam0 = 0.27

#mu = 3./7 * np.exp(-2/63. * np.log(omegam0))



### Get the power spectrum from an external file


def power_spectrum():
    
    monfichier = open('matterpower.dat','r')
    matterpower = monfichier.readlines()
    monfichier.close()
    result = np.array([[0,0]])
    for ligne in matterpower:
        part = ligne.split('    ')
        nb = np.array([[float(part[1]),float(part[2])]])
        result = np.concatenate((result,nb))

    return (result[1:,:])
    
powerspectrum = power_spectrum()
N = powerspectrum.size/2    #powerspectrum is a 2 column vector

###Growth factor

def E2(z) : return omegam0*(1+z)**3 + omegalambda0

def D(z) :
    omegam = omegam0 / E2(z) * (1+z)**3 
    omegalambda = omegalambda0 / E2(z)
    return 5/2. * omegam / ((1+z) * (pow(omegam, 4/7.) - omegalambda + (1 + omegam/2)*(1 + omegalambda/70)))



### function which interpolates the power spectrum

def P(k, z) :
    if k < powerspectrum[0,0] : print 'Points missing on the left : k_asked = {}, k_min = {}'.format(k, powerspectrum[0,0])
    elif k > powerspectrum[-1,0] : print 'Points missing on the right : k_asked = {}, k_max = {}'.format(k, powerspectrum[-1,0])
    
    return np.interp(k, powerspectrum[:,0], powerspectrum[:,1]) * (D(z)/D(0))**2
   


#### Bispectrum

def norme2(k1, k2, u12) :
    
    tmp = k1**2 + k2**2 + 2*k1*k2*u12
    if tmp < 0. : print 'oops ! Negative norm2 : {} ! k1 = {}, k2 = {}'.format(tmp, k1, k2)
    return np.sqrt(np.abs(tmp))
    
def norme3(k1, k2, k3, u12, u13, u23) :
    
    tmp = k1**2 + k2**2 + k3**2 + 2*k1*k2*u12 + 2*k1*k3*u13 + 2*k2*k3*u23
    if tmp < 0. : print 'oops ! Negative norm3 : {} !'.format(tmp)
    return np.sqrt(np.abs(tmp))


def F2(k1, k2, u12) : return 5./7 + 1/2.*(k1/k2 + k2/k1)*u12 + 2./7*u12**2
    

def BPT(k1, k2, k3, u12, u13, u23, z) :

    return (2*F2(k1, k2, u12)*P(k1, z)*P(k2, z) + 2*F2(k1, k3, u13)*P(k1, z)*P(k3, z) + 2*F2(k2, k3, u23)*P(k2, z)*P(k3, z))



### Trispectrum
##return T(k1, -k1, k2, -k2)

def G2(k1, k2, u12) : return 3./7 + 1/2.*(k1/k2 + k2/k1)*u12 + 4./7*u12**2
    


def F3(k1, k2, k3, u12, u13, u23) :
    n123 = norme3(k1, k2, k3, u12, u13, u23)**2
    n12 = norme2(k1, k2, u12)**2
    n23 = norme2(k2, k3, u23)**2
    
    
    if n12 == 0. :                  #F3(k1, -k1, k2) is frequently used
        terme1 = 7.*(1 + k2/k1*u12 + k3/k1*u13)
        terme2 = n123*(k1*k2*u12 + k1*k3*u13)/(k1**2 * n23)
        return 1/18.*(terme1 * F2(k2, k3, u23) + terme2 * G2(k2, k3, u23))
        
    elif n23 == 0. :
        terme3 = 7.*(k1**2 + k2**2 + 2*k1*k2*u12 + k1*k3*u13 + k2*k3*u23)/(n12) + n123*(k1*k3*u13 + k2*k3*u23)/(n12*k3**2)
        return 1/18.*(terme3 * G2(k1, k2, u12))

    else :
        terme1 = 7.*(1 + k2/k1*u12 + k3/k1*u13)
        terme2 = n123*(k1*k2*u12 + k1*k3*u13)/(k1**2 * n23)
        terme3 = 7.*(k1**2 + k2**2 + 2*k1*k2*u12 + k1*k3*u13 + k2*k3*u23)/(n12) + n123*(k1*k3*u13 + k2*k3*u23)/(n12*k3**2)
        return 1/18.*(terme1 * F2(k2, k3, u23) + terme2 * G2(k2, k3, u23) + terme3 * G2(k1, k2, u12))
    
    
    
    
def F3s(k1, k2, k3, u12, u13, u23):
    return 1/6.*(F3(k1, k2, k3, u12, u13, u23) + F3(k1, k3, k2, u13, u12, u23) + F3(k2, k1, k3, u12, u23, u13) + F3(k3, k1, k2, u13, u23, u12) + F3(k3, k2, k1, u23, u13, u12) + F3(k2, k3, k1, u23, u12, u13))



def TPT(k1, k2, u12, z):
    n12 = norme2(k1, k2, u12)
    m12 = norme2(k1, k2, -u12)
    
    return ( 4*P(n12, z)*(F2(k1, n12, -(k1**2 + k1*k2*u12)/(k1*n12)) * P(k1, z) + F2(k2, n12, -(k2**2 + k1*k2*u12)/(k2*n12)) * P(k2, z))**2
           + 4*P(m12, z)*(F2(k1, m12, (k1*k2*u12 - k1**2)/(k1*m12)) * P(k1, z) + F2(k2, m12, (k1*k2*u12 - k2**2)/(k2*m12)) * P(k2, z))**2
           + 12*(F3s(k1, k1, k2, -1., u12, -u12) * P(k1, z)**2 * P(k2, z) + F3s(k1, k2, k2, u12, -u12, -1.) * P(k1, z) * P(k2, z)**2) )


# -*- coding: utf-8 -*-
"""
Created on Thu Apr  2 15:00:25 2015

@author: adrienkuntz
"""

import numpy as np
cimport numpy as np
from scipy import integrate
from scipy import optimize
cimport cython_gsl as gsl
from Params import *
#from scipy.special import sici
#from scipy.optimize import fsolve
#import matplotlib.pyplot as plt
#import math





### comoving distance as a function of redshift

def E(z): return np.sqrt(omegam0 * (1 + z)**3 + omegalambda0)
    
def chi(z) : return cl/H0 * integrate.quad(lambda x : 1/E(x), 0., z, epsrel = 1e-6, epsabs = 0)[0]
    
def chiv(z) : return np.vectorize(chi)(z)
    
def redshift(chi_): return optimize.brentq(lambda x : chiv(x) - chi_, 0., zCMB)

def redshiftv(chi) : return np.array(map(redshift, chi))
    
 
cdef double chiCMB = chi(zCMB)


#define the kernels ; for convenience I take an analytical dN/dz

def Wk(chi) : return 3./2 * H0**2 / cl**2 * omegam0 * chi * (chiCMB - chi) * (1 + redshift(chi)) / chiCMB

def dNdz(z) : return alpha / z0 * np.exp(lambda_ * np.log(z/z0)) * np.exp(-np.exp(beta * np.log(z/z0)))
    
intgl = integrate.quad(dNdz, 0., zCMB, epsabs = 0, epsrel = 1e-6)[0]

def Wg(chi_) :
    za = redshift(chi_)
    f = chi_ * integrate.quad(lambda z : dNdz(z) / intgl * (chi(z)-chi_) / chi(z), za, zCMB, epsabs=0, epsrel=1e-3)[0]
    return dNdz(za) / intgl * H0 / cl * E(za) + 3/2. * omegam0 * (H0/cl)**2 * (1+za) * (5*sn-2) * f
    
    
cdef double [::1] chitab = np.linspace(0., chiCMB, chinum)
cdef int i

cdef double [::1] Wktab = np.zeros(chinum)
cdef double [::1] Wgtab = np.zeros(chinum)
cdef double [::1] redshifttab = np.zeros(chinum)

for i in range(chinum) :
    redshifttab[i] = redshift(chitab[i])
       

for i in range(chinum) :
    Wktab[i] = Wk(chitab[i])
    Wgtab[i] = Wg(chitab[i])
    redshifttab[i] = redshift(chitab[i])
    
Wkfile = open('Wk.dat', 'w')
for i in range(chinum) :
    Wkfile.write('{} {}\n'.format(chitab[i], Wktab[i]))
Wkfile.close()

Wgfile = open('Wg.dat', 'w')
for i in range(chinum) :
    Wgfile.write('{} {}\n'.format(chitab[i], Wgtab[i]))
Wgfile.close()


cdef gsl.gsl_interp_accel *accWk = gsl.gsl_interp_accel_alloc()
cdef gsl.gsl_interp *interpWk = gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, chinum) 
gsl.gsl_interp_init(interpWk, &chitab[0], &Wktab[0], chinum)

cdef gsl.gsl_interp_accel *accWg = gsl.gsl_interp_accel_alloc()
cdef gsl.gsl_interp *interpWg = gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, chinum) 
gsl.gsl_interp_init(interpWg, &chitab[0], &Wgtab[0], chinum)

cdef gsl.gsl_interp_accel *accz = gsl.gsl_interp_accel_alloc()
cdef gsl.gsl_interp *interpz = gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, chinum) 
gsl.gsl_interp_init(interpz, &chitab[0], &redshifttab[0], chinum)


i=0
monfichier = open('Wk.dat', 'r')
iterable = monfichier.readlines()
monfichier.close()
for ligne in iterable :
    part = ligne.split(' ')
    Wktab[i] = float(part[1])
    i = i + 1
i = 0
monfichier = open('Wg.dat', 'r')
iterable = monfichier.readlines()
monfichier.close()
for ligne in iterable :
    part = ligne.split(' ')
    Wgtab[i] = float(part[1])
    i = i + 1

cpdef double Wki(double chi) :
    return gsl.gsl_interp_eval(interpWk, &chitab[0], &Wktab[0], chi, accWk)
    
cpdef double Wgi(double chi) :
    return gsl.gsl_interp_eval(interpWg, &chitab[0], &Wgtab[0], chi, accWg)

cpdef double redshifti(double chi) :
    return gsl.gsl_interp_eval(interpz, &chitab[0], &redshifttab[0], chi, accz)


# -*- coding: utf-8 -*-
"""
Created on Tue Apr  7 12:40:46 2015

@author: adrienkuntz
"""

cpdef double Wki(double chi)
cpdef double Wgi(double chi)
cpdef double redshifti(double chi)
cdef double chiCMB


# -*- coding: utf-8 -*-
"""
Created on Mon Mar 30 17:06:01 2015

@author: adrienkuntz
"""

##### Calculation of the bispectrum and the trispectrum in perturbation theory
##### The trispectrum is evaluated in points (k1, -k1, k2, -k2) for the purpose of the study
##### In this program uij means cos(theta(i,j)) with theta(i,j) the angle between ki and kj

import numpy as np
cimport numpy as np
cimport cython_gsl as gsl
from Params import *
import kernels as knl
cdef extern from "math.h" :
    double pow(double, double)

#@cython.boundscheck(False)
#@cython.wraparound(False)




### Get the power spectrum from an external file



    
monfichier = open('matterpower.dat','r')
matterpower = monfichier.readlines()
monfichier.close()
result = np.array([[0,0]])
for ligne in matterpower:
    part = ligne.split('    ')
    nb = np.array([[float(part[1]),float(part[2])]])
    result = np.concatenate((result,nb))


powerspectrum = result[1:,:]
cpdef size_t N = powerspectrum.size/2

cdef double [::1] ktab = powerspectrum[:,0].copy()
cdef double [::1] Ptab = powerspectrum[:,1].copy()


cdef double *kptr
kptr = &ktab[0]
cdef double *Pptr
Pptr = &Ptab[0]


###Growth factor

cpdef double E2(double z) : return omegam0*(1+z)**3 + omegalambda0

cpdef double D(double z) :
    cdef double omegam = omegam0 / E2(z) * (1+z)**3 
    cdef double omegalambda = omegalambda0 / E2(z)
    return 5/2. * omegam / ((1+z) * (pow(omegam, 4/7.) - omegalambda + (1 + omegam/2)*(1 + omegalambda/70)))




### function which interpolates the power spectrum

cdef gsl.gsl_interp_accel *acc = gsl.gsl_interp_accel_alloc ()
cdef gsl.gsl_interp *interp = gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, N) 
gsl.gsl_interp_init(interp, &ktab[0], &Ptab[0], N)


cpdef double P(double k, double z) :
    
    if k < ktab[0] :
        if k == 0 : return 0
        else :
           # print 'Points missing on the left for P : k_asked = {}, k_min = {}'.format(k, ktab[0])
            return Ptab[0] * (D(z)/D(0))**2
    elif k > ktab[N-1] :
        #print 'Points missing on the right for P : k_asked = {}, k_max = {}'.format(k, ktab[N-1])
        return Ptab[N-1] * (D(z)/D(0))**2
    else :
        return gsl.gsl_interp_eval(interp, &ktab[0], &Ptab[0], k, acc) * (D(z)/D(0))**2




#### Bispectrum

cdef double norme2(double k1, double k2, double u12) :
    
    cdef double tmp = k1**2 + k2**2 + 2*k1*k2*u12
    #if tmp < 0. : print 'oops ! Negative norm2 : {} ! k1 = {}, k2 = {}'.format(tmp, k1, k2)    #sometimes a tiny negative norm can happen, due to residual errors
    return gsl.sqrt(gsl.fabs(tmp))
    
cdef double norme3(double k1, double k2, double k3, double u12, double u13, double u23) :
    
    cdef double tmp = k1**2 + k2**2 + k3**2 + 2*k1*k2*u12 + 2*k1*k3*u13 + 2*k2*k3*u23
    #if tmp < 0. : print 'oops ! Negative norm3 : {} !'.format(tmp)
    return gsl.sqrt(gsl.fabs(tmp))


cdef double F2(double k1, double k2, double u12) :
    if (k1 == 0.) or (k2 == 0.) :
        return 0.
    else :
        return 5./7 + 1/2.*(k1/k2 + k2/k1)*u12 + 2./7*u12**2
    

cdef double BPT(double k1, double k2, double k3, double u12, double u13, double u23, double z) :
    
    return (2*F2(k1, k2, u12)*P(k1, z)*P(k2, z) + 2*F2(k1, k3, u13)*P(k1, z)*P(k3, z) + 2*F2(k2, k3, u23)*P(k2, z)*P(k3, z))



### Trispectrum
##return T(k1, -k1, k2, -k2)

cdef double G2(double k1, double k2, double u12) : return 3./7 + 1/2.*(k1/k2 + k2/k1)*u12 + 4./7*u12**2
    


cdef double F3(double k1, double k2, double k3, double u12, double u13, double u23) :
    cdef double n123 = norme3(k1, k2, k3, u12, u13, u23)**2
    cdef double n12 = norme2(k1, k2, u12)**2
    cdef double n23 = norme2(k2, k3, u23)**2
    cdef double terme1, terme2, terme3
    
    if (n12 == 0.) and (n23 == 0.) :
        return 0.
    
    elif n12 == 0. :                  #F3(k1, -k1, k2) is frequently used
        terme1 = 7.*(1 + k2/k1*u12 + k3/k1*u13)
        terme2 = n123*(k1*k2*u12 + k1*k3*u13)/(k1**2 * n23)
        return 1/18.*(terme1 * F2(k2, k3, u23) + terme2 * G2(k2, k3, u23))
        
    elif n23 == 0. :
        terme3 = 7.*(k1**2 + k2**2 + 2*k1*k2*u12 + k1*k3*u13 + k2*k3*u23)/(n12) + n123*(k1*k3*u13 + k2*k3*u23)/(n12*k3**2)
        return 1/18.*(terme3 * G2(k1, k2, u12))

    else :
        terme1 = 7.*(1 + k2/k1*u12 + k3/k1*u13)
        terme2 = n123*(k1*k2*u12 + k1*k3*u13)/(k1**2 * n23)
        terme3 = 7.*(k1**2 + k2**2 + 2*k1*k2*u12 + k1*k3*u13 + k2*k3*u23)/(n12) + n123*(k1*k3*u13 + k2*k3*u23)/(n12*k3**2)
        return 1/18.*(terme1 * F2(k2, k3, u23) + terme2 * G2(k2, k3, u23) + terme3 * G2(k1, k2, u12))
    
    
    
    
cdef double F3s(double k1, double k2, double k3, double u12, double u13, double u23):
    return 1/6.*(F3(k1, k2, k3, u12, u13, u23) + F3(k1, k3, k2, u13, u12, u23) + F3(k2, k1, k3, u12, u23, u13) + F3(k3, k1, k2, u13, u23, u12) + F3(k3, k2, k1, u23, u13, u12) + F3(k2, k3, k1, u23, u12, u13))



cdef double TPT(double k1, double k2, double u12, double z):
    cdef double n12 = norme2(k1, k2, u12)
    cdef double m12 = norme2(k1, k2, -u12)
    cdef double u1_n12, u2_n12, u1_m12, u2_m12
    
    if n12 == 0 :
        u1_n12 = 0.
        u2_n12 = 0.
    else :
        u1_n12 = (k1**2 + k1*k2*u12)/(k1 * n12)      #angle between k1 and k1+k2, etc
        u2_n12 = (k2**2 + k1*k2*u12)/(k2 * n12)
        
    if m12 == 0 :
        u1_m12 = 0.
        u2_m12 = 0.
    else :
        u1_m12 = (k1**2 - k1*k2*u12)/(k1 * m12)     
        u2_m12 = (-k2**2 + k1*k2*u12)/(k2 * m12)

    
    return ( 4*P(n12, z)*(F2(k1, n12, -u1_n12) * P(k1, z) + F2(k2, n12, -u2_n12) * P(k2, z))**2
           + 4*P(m12, z)*(F2(k1, m12, -u1_m12) * P(k1, z) + F2(k2, m12, u2_m12) * P(k2, z))**2
           + 12*(F3s(k1, k1, k2, -1., u12, -u12) * P(k1, z)**2 * P(k2, z) + F3s(k1, k2, k2, u12, -u12, -1.) * P(k1, z) * P(k2, z)**2) )


# -*- coding: utf-8 -*-
"""
Created on Wed Apr  1 14:04:31 2015

@author: adrienkuntz
"""

cpdef size_t N
cdef double *kptr
cdef double *Pptr

cpdef double P(double k, double z)
cdef double norme2(double k1, double k2, double u12)
cdef double norme3(double k1, double k2, double k3, double u12, double u13, double u23)
cdef double F2(double k1, double k2, double u12)
cdef double BPT(double k1, double k2, double k3, double u12, double u13, double u23, double z)
cdef double G2(double k1, double k2, double u12)
cdef double F3(double k1, double k2, double k3, double u12, double u13, double u23)
cdef double F3s(double k1, double k2, double k3, double u12, double u13, double u23)
cdef double TPT(double k1, double k2, double u12, double z)


# -*- coding: utf-8 -*-
"""
Created on Fri Mar 13 13:36:51 2015

@author: adrienkuntz
"""

import numpy as np
from scipy import integrate
from scipy import optimize
from Params import *
#from scipy.special import sici
#from scipy.optimize import fsolve
#import matplotlib.pyplot as plt
#import math



### comoving distance as a function of redshift

def E(z): return np.sqrt(omegam0 * (1 + z)**3 + omegalambda0)
    
def chi(z) : return cl/H0 * integrate.quad(lambda x : 1/E(x), 0., z, epsrel = 1e-6, epsabs = 0)[0]
    
def chiv(z) : return np.vectorize(chi)(z)
    
def redshift(chi_): return optimize.brentq(lambda x : chiv(x) - chi_, 0., zCMB)

def redshiftv(chi) : return np.array(map(redshift, chi))
    
 
chiCMB = chi(zCMB)


#define the kernels ; for convenience I take an analytical dN/dz

def Wk(chi) : return 3./2 * H0**2 / cl**2 * omegam0 * chi * (chiCMB - chi) * (1 + redshift(chi)) / chiCMB

def dNdz(z) : return alpha / z0 * np.exp(lambda_ * np.log(z/z0)) * np.exp(-np.exp(beta * np.log(z/z0)))
    
intgl = integrate.quad(dNdz, 0., zCMB, epsabs = 0, epsrel = 1e-6)[0]

def Wg(chi_) :
    za = redshift(chi_)
    f = chi_ * integrate.quad(lambda z : dNdz(z) / intgl * (chi(z)-chi_) / chi(z), za, zCMB, epsabs=0, epsrel=1e-3)[0]
    return dNdz(za) / intgl * H0 / cl * E(za) + 3/2. * omegam0 * (H0/cl)**2 * (1+za) * (5*sn-2) * f


# -*- coding: utf-8 -*-
"""
Created on Fri Apr 17 10:28:57 2015

@author: adrienkuntz
"""
import numpy as np


### In kernelsc.pyx
h = 0.679
H0 = 100e3                                           #hubble constant in m/s/Mpc * h
cl = 3e8                                             #speed of light in m/s
omegalambda0 = 0.7                                 #omega parameters
omegam0 = 1 - omegalambda0
zCMB = 1090                                          #redshift of the CMB
alpha = 2.2                                         #parameters for dn/dz found by fitting the real dN/dz ; see fit_dNdz.py
lambda_ = 0.781                                      #parameters for dn/dz
beta = 3.36                                           #parameters for dn/dz
z0 = 0.973                                           #parameters for dn/dz (version CFHTLenS)
sn = 1.                                              #parameter for the magnification bias

chinum = 3000                                       #size for the storage of the values of Wk and Wg



### in M.py

G = 6.67e-11                                         #Gravitational constant
delta_sc = 1.686                                     #critical density of spherical collapse
rhobar = 3*H0**2 / (8*np.pi*G) * omegam0 / 2e30 * (3.09e22)    #average comoving density of matter of the universe (Solar mass/h * (h/Mpc)**3)
q = 0.75
p = 0.3                                              #parameters for the Sheth & Tormen function
delta_nl = np.exp(1/3.*np.log(340))                  #ratio between the virial radius and the initial radius of a halo in the spherical collapse model
a2 = -17/21.
a3 = 341/567.

mnum = 1000
mmin = 1e-6
xmax = 10.                                            #Maximum x for the integration of WTH(x)
nspectral = 0.96                                             #Spectral index for the initial power spectrum


knum = 300                                           #number of points for the storage of the values of M in arrays
kmin_log = -3
kmax_log = 3

zmin = 0
zmax = 2.5
znum = 30
ztab = np.linspace(zmin, zmax, znum)


# -*- coding: utf-8 -*-
"""
Created on Tue Apr  7 11:41:13 2015

@author: adrienkuntz
"""

import numpy as np
cimport numpy as np
cimport cython_gsl as gsl
cimport Trispectrum_PTc as ptc
from Params import *
from libc.stdlib cimport malloc, free


cdef double [::1] ktab = np.logspace(kmin_log, kmax_log, knum)
cdef double [:] ztabc = ztab.copy()
cdef double zmax = zmax


#function which performs a search in a sorted array in c

cdef int search(double [:] tab, int length, double value) :
    cdef int i = 0, result = -1
    while (i < length) and (result == -1) :
        if tab[i] >= value :
            result = i
        i += 1
    return result



###Define the arrays containing the values of M






cdef double [:,::1] MT11_k1 = np.zeros((znum, knum))
cdef double [:,::1] MT21_k1 = np.zeros((znum, knum))
cdef double [:,::1] MT31_k1 = np.zeros((znum, knum))
cdef double [:,:,::1] MT12_k1k2 = np.zeros((znum, knum, knum), order = 'C')
cdef double [:,:,::1] MT22_k1k2 = np.zeros((znum, knum, knum), order = 'C')
cdef double [:,:,::1] MT04_k1k1k2k2 = np.zeros((znum, knum, knum), order = 'C')
cdef double [:,:,::1] MT13_k1k1k2 = np.zeros((znum, knum, knum), order = 'C')
cdef double [:,::1] MT02_k1k1 = np.zeros((znum, knum))



##definitions for the interpolation. The values of z are indexed in an array (very simple interpolation since there is no function interp3d available)

cdef gsl.gsl_interp_accel **accM11_k1 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.gsl_interp **interpM11_k1 = <gsl.gsl_interp **> malloc(znum * sizeof(gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, knum)))

cdef gsl.gsl_interp_accel **accM21_k1 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.gsl_interp **interpM21_k1 = <gsl.gsl_interp **> malloc(znum * sizeof(gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, knum)))

cdef gsl.gsl_interp_accel **accM31_k1 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.gsl_interp **interpM31_k1 = <gsl.gsl_interp **> malloc(znum * sizeof(gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, knum)))

cdef gsl.gsl_interp_accel **accM02_k1k1 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.gsl_interp **interpM02_k1k1 = <gsl.gsl_interp **> malloc(znum * sizeof(gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, knum)))



cdef gsl.gsl_interp_accel **xaccM12_k1k2 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.gsl_interp_accel **yaccM12_k1k2 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.interp2d **interpM12_k1k2 = <gsl.interp2d **> malloc(znum * sizeof(gsl.interp2d_alloc(gsl.interp2d_bicubic, knum, knum)))

cdef gsl.gsl_interp_accel **xaccM22_k1k2 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.gsl_interp_accel **yaccM22_k1k2 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.interp2d **interpM22_k1k2 = <gsl.interp2d **> malloc(znum * sizeof(gsl.interp2d_alloc(gsl.interp2d_bicubic, knum, knum)))

cdef gsl.gsl_interp_accel **xaccM04_k1k1k2k2 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.gsl_interp_accel **yaccM04_k1k1k2k2 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.interp2d **interpM04_k1k1k2k2 = <gsl.interp2d **> malloc(znum * sizeof(gsl.interp2d_alloc(gsl.interp2d_bicubic, knum, knum)))

cdef gsl.gsl_interp_accel **xaccM13_k1k1k2 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.gsl_interp_accel **yaccM13_k1k1k2 = <gsl.gsl_interp_accel **> malloc(znum * sizeof(gsl.gsl_interp_accel*))
cdef gsl.interp2d **interpM13_k1k1k2 = <gsl.interp2d **> malloc(znum * sizeof(gsl.interp2d_alloc(gsl.interp2d_bicubic, knum, knum)))










cdef int l
cdef int i = 0
cdef int j = 0

for l in range(znum) :  
    
    ###Get M from the files
    
    monfichier = open('M11_k1_{}.dat'.format(ztabc[l]), 'r')
    iterable = monfichier.readlines()
    monfichier.close()
    for ligne in iterable :
        part = ligne.split(' ')
        MT11_k1[l, i] = float(part[1])
        i = i + 1
    i = 0
    
    
        
    monfichier = open('M21_k1_{}.dat'.format(ztabc[l]), 'r')
    iterable = monfichier.readlines()
    monfichier.close()
    for ligne in iterable :
        part = ligne.split(' ')
        MT21_k1[l, i] = float(part[1])
        i = i + 1
    i = 0
    
    monfichier = open('M31_k1_{}.dat'.format(ztabc[l]), 'r')
    iterable = monfichier.readlines()
    monfichier.close()
    for ligne in iterable :
        part = ligne.split(' ')
        MT31_k1[l, i] = float(part[1])
        i = i + 1
    i = 0
    
       
    
    
    
    monfichier = open('M12_k1k2_{}.dat'.format(ztabc[l]), 'r')
    iterable = monfichier.readlines()
    monfichier.close()
    for ligne in iterable :
        part = ligne.split(' ')
        MT12_k1k2[l, i, j] = float(part[2])
        j = j + 1
        if j == knum :
            i = i + 1
            j = 0
    i = 0
    j = 0
         
         
    
    monfichier = open('M22_k1k2_{}.dat'.format(ztabc[l]), 'r')
    iterable = monfichier.readlines()
    monfichier.close()
    for ligne in iterable :
        part = ligne.split(' ')
        MT22_k1k2[l, i, j] = float(part[2])
        j = j + 1
        if j == knum :
            i = i + 1
            j = 0
    i = 0
    j = 0
            
            
    
    monfichier = open('M04_k1k1k2k2_{}.dat'.format(ztabc[l]), 'r')
    iterable = monfichier.readlines()
    monfichier.close()
    for ligne in iterable :
        part = ligne.split(' ')
        MT04_k1k1k2k2[l, i, j] = float(part[2])
        j = j + 1
        if j == knum :
            i = i + 1
            j = 0
    i = 0
    j = 0
    
    
    
    monfichier = open('M13_k1k1k2_{}.dat'.format(ztabc[l]), 'r')
    iterable = monfichier.readlines()
    monfichier.close()
    for ligne in iterable :
        part = ligne.split(' ')
        MT13_k1k1k2[l, i, j] = float(part[2])
        j = j + 1
        if j == knum :
            i = i + 1
            j = 0
    i = 0
    j = 0
    
    
    
    
    monfichier = open('M02_k1k1_{}.dat'.format(ztabc[l]), 'r')
    iterable = monfichier.readlines()
    monfichier.close()
    for ligne in iterable :
        part = ligne.split(' ')
        MT02_k1k1[l, i] = float(part[1])
        i = i + 1
    i = 0
    
    
    
    
    ###initialize the interpolation
    
    accM11_k1[l] = gsl.gsl_interp_accel_alloc()
    interpM11_k1[l] = gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, knum)
    gsl.gsl_interp_init(interpM11_k1[l], &ktab[0], &MT11_k1[l, 0], knum)
    
    accM21_k1[l] = gsl.gsl_interp_accel_alloc()
    interpM21_k1[l] = gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, knum)
    gsl.gsl_interp_init(interpM21_k1[l], &ktab[0], &MT21_k1[l, 0], knum)
    
    accM31_k1[l] = gsl.gsl_interp_accel_alloc()
    interpM31_k1[l] = gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, knum)
    gsl.gsl_interp_init(interpM31_k1[l], &ktab[0], &MT31_k1[l, 0], knum)
    
    accM02_k1k1[l] = gsl.gsl_interp_accel_alloc()
    interpM02_k1k1[l] = gsl.gsl_interp_alloc(gsl.gsl_interp_cspline, knum)
    gsl.gsl_interp_init(interpM02_k1k1[l], &ktab[0], &MT02_k1k1[l, 0], knum)
   
  
    
    
    xaccM12_k1k2[l] = gsl.gsl_interp_accel_alloc()
    yaccM12_k1k2[l] = gsl.gsl_interp_accel_alloc()
    interpM12_k1k2[l] = gsl.interp2d_alloc(gsl.interp2d_bicubic, knum, knum)
    gsl.interp2d_init(interpM12_k1k2[l], &ktab[0], &ktab[0], &MT12_k1k2[l, 0, 0], knum, knum)
    
    xaccM22_k1k2[l] = gsl.gsl_interp_accel_alloc()
    yaccM22_k1k2[l] = gsl.gsl_interp_accel_alloc()
    interpM22_k1k2[l] = gsl.interp2d_alloc(gsl.interp2d_bicubic, knum, knum)
    gsl.interp2d_init(interpM22_k1k2[l], &ktab[0], &ktab[0], &MT22_k1k2[l, 0, 0], knum, knum)
    
    xaccM04_k1k1k2k2[l] = gsl.gsl_interp_accel_alloc()
    yaccM04_k1k1k2k2[l] = gsl.gsl_interp_accel_alloc()
    interpM04_k1k1k2k2[l] = gsl.interp2d_alloc(gsl.interp2d_bicubic, knum, knum)
    gsl.interp2d_init(interpM04_k1k1k2k2[l], &ktab[0], &ktab[0], &MT04_k1k1k2k2[l, 0, 0], knum, knum)
    
    xaccM13_k1k1k2[l] = gsl.gsl_interp_accel_alloc()
    yaccM13_k1k1k2[l] = gsl.gsl_interp_accel_alloc()
    interpM13_k1k1k2[l] = gsl.interp2d_alloc(gsl.interp2d_bicubic, knum, knum)
    gsl.interp2d_init(interpM13_k1k1k2[l], &ktab[0], &ktab[0], &MT13_k1k1k2[l, 0, 0], knum, knum)






##interpolation functions
##To avoid too much verbose it does not display a message when the asked point is outside the range of ktab. Only P does this.

## 1D ###########################################################################################

cpdef double M11_k1(double k1, double z):
    cdef int i = 0
    cdef double left, right, t
    
    if z >= zmax :
        if k1 < ktab[0] :
            return MT11_k1[znum-1, 0]
        elif k1 > ktab[knum-1] :
            return MT11_k1[znum-1, knum-1]
        else :
            return gsl.gsl_interp_eval(interpM11_k1[znum-1], &ktab[0], &MT11_k1[znum-1, 0], k1, accM11_k1[znum-1])
            
    else :
        i = search(ztabc, znum, z)
        t = (z - ztabc[i])/(ztabc[i-1] - ztabc[i])
        
        if k1 < ktab[0] :
            left = MT11_k1[i-1, 0]
            right = MT11_k1[i, 0]
        elif k1 > ktab[knum-1] :
            left = MT11_k1[i-1, knum-1]
            right = MT11_k1[i, knum-1]
        else :
            left = gsl.gsl_interp_eval(interpM11_k1[i-1], &ktab[0], &MT11_k1[i-1, 0], k1, accM11_k1[i-1])
            right = gsl.gsl_interp_eval(interpM11_k1[i], &ktab[0], &MT11_k1[i, 0], k1, accM11_k1[i])
            
        return t*left + (1-t)*right
        


cpdef double M21_k1(double k1, double z):
    cdef int i = 0
    cdef double left, right, t
    
    if z >= zmax :
        if k1 < ktab[0] :
            return MT21_k1[znum-1, 0]
        elif k1 > ktab[knum-1] :
            return MT21_k1[znum-1, knum-1]
        else :
            return gsl.gsl_interp_eval(interpM21_k1[znum-1], &ktab[0], &MT21_k1[znum-1, 0], k1, accM21_k1[znum-1])
            
    else :
        i = search(ztabc, znum, z)
        t = (z - ztabc[i])/(ztabc[i-1] - ztabc[i])
        
        if k1 < ktab[0] :
            left = MT21_k1[i-1, 0]
            right = MT21_k1[i, 0]
        elif k1 > ktab[knum-1] :
            left = MT21_k1[i-1, knum-1]
            right = MT21_k1[i, knum-1]
        else :
            left = gsl.gsl_interp_eval(interpM21_k1[i-1], &ktab[0], &MT21_k1[i-1, 0], k1, accM21_k1[i-1])
            right = gsl.gsl_interp_eval(interpM21_k1[i], &ktab[0], &MT21_k1[i, 0], k1, accM21_k1[i])
            
        return t*left + (1-t)*right
        

cpdef double M31_k1(double k1, double z):
    cdef int i = 0
    cdef double left, right, t
    
    if z >= zmax :
        if k1 < ktab[0] :
            return MT31_k1[znum-1, 0]
        elif k1 > ktab[knum-1] :
            return MT31_k1[znum-1, knum-1]
        else :
            return gsl.gsl_interp_eval(interpM31_k1[znum-1], &ktab[0], &MT31_k1[znum-1, 0], k1, accM31_k1[znum-1])
            
    else :
        i = search(ztabc, znum, z)
        t = (z - ztabc[i])/(ztabc[i-1] - ztabc[i])
        
        if k1 < ktab[0] :
            left = MT31_k1[i-1, 0]
            right = MT31_k1[i, 0]
        elif k1 > ktab[knum-1] :
            left = MT31_k1[i-1, knum-1]
            right = MT31_k1[i, knum-1]
        else :
            left = gsl.gsl_interp_eval(interpM31_k1[i-1], &ktab[0], &MT31_k1[i-1, 0], k1, accM31_k1[i-1])
            right = gsl.gsl_interp_eval(interpM31_k1[i], &ktab[0], &MT31_k1[i, 0], k1, accM31_k1[i])
            
        return t*left + (1-t)*right




cpdef double M02_k1k1(double k1, double z):
    cdef int i = 0
    cdef double left, right, t
    
    if z >= zmax :
        if k1 < ktab[0] :
            return MT02_k1k1[znum-1, 0]
        elif k1 > ktab[knum-1] :
            return MT02_k1k1[znum-1, knum-1]
        else :
            return gsl.gsl_interp_eval(interpM02_k1k1[znum-1], &ktab[0], &MT02_k1k1[znum-1, 0], k1, accM02_k1k1[znum-1])
            
    else :
        i = search(ztabc, znum, z)
        t = (z - ztabc[i])/(ztabc[i-1] - ztabc[i])
        
        if k1 < ktab[0] :
            left = MT02_k1k1[i-1, 0]
            right = MT02_k1k1[i, 0]
        elif k1 > ktab[knum-1] :
            left = MT02_k1k1[i-1, knum-1]
            right = MT02_k1k1[i, knum-1]
        else :
            left = gsl.gsl_interp_eval(interpM02_k1k1[i-1], &ktab[0], &MT02_k1k1[i-1, 0], k1, accM02_k1k1[i-1])
            right = gsl.gsl_interp_eval(interpM02_k1k1[i], &ktab[0], &MT02_k1k1[i, 0], k1, accM02_k1k1[i])
            
        return t*left + (1-t)*right




   
## 2D ################################################################################################### 
  
  



cpdef double M12_k1k2(double k1, double k2, double z):
    
    if z >= zmax :
        return gsl.interp2d_eval_no_boundary_check(interpM12_k1k2[znum-1], &ktab[0], &ktab[0], &MT12_k1k2[znum-1,0,0], k1, k2, xaccM12_k1k2[znum-1], yaccM12_k1k2[znum-1])

    else :
        i = search(ztabc, znum, z)
        t = (z - ztabc[i])/(ztabc[i-1] - ztabc[i])
    
        left = gsl.interp2d_eval_no_boundary_check(interpM12_k1k2[i-1], &ktab[0], &ktab[0], &MT12_k1k2[i-1,0,0], k1, k2, xaccM12_k1k2[i-1], yaccM12_k1k2[i-1])
        right = gsl.interp2d_eval_no_boundary_check(interpM12_k1k2[i], &ktab[0], &ktab[0], &MT12_k1k2[i,0,0], k1, k2, xaccM12_k1k2[i], yaccM12_k1k2[i])
            
        return t*left + (1-t)*right


cpdef double M22_k1k2(double k1, double k2, double z):
    
    if z >= zmax :
        return gsl.interp2d_eval_no_boundary_check(interpM22_k1k2[znum-1], &ktab[0], &ktab[0], &MT22_k1k2[znum-1,0,0], k1, k2, xaccM22_k1k2[znum-1], yaccM22_k1k2[znum-1])

    else :
        i = search(ztabc, znum, z)
        t = (z - ztabc[i])/(ztabc[i-1] - ztabc[i])
    
        left = gsl.interp2d_eval_no_boundary_check(interpM22_k1k2[i-1], &ktab[0], &ktab[0], &MT22_k1k2[i-1,0,0], k1, k2, xaccM22_k1k2[i-1], yaccM22_k1k2[i-1])
        right = gsl.interp2d_eval_no_boundary_check(interpM22_k1k2[i], &ktab[0], &ktab[0], &MT22_k1k2[i,0,0], k1, k2, xaccM22_k1k2[i], yaccM22_k1k2[i])
            
        return t*left + (1-t)*right


cpdef double M04_k1k1k2k2(double k1, double k2, double z):
    
    if z >= zmax :
        return gsl.interp2d_eval_no_boundary_check(interpM04_k1k1k2k2[znum-1], &ktab[0], &ktab[0], &MT04_k1k1k2k2[znum-1,0,0], k1, k2, xaccM04_k1k1k2k2[znum-1], yaccM04_k1k1k2k2[znum-1])

    else :
        i = search(ztabc, znum, z)
        t = (z - ztabc[i])/(ztabc[i-1] - ztabc[i])
    
        left = gsl.interp2d_eval_no_boundary_check(interpM04_k1k1k2k2[i-1], &ktab[0], &ktab[0], &MT04_k1k1k2k2[i-1,0,0], k1, k2, xaccM04_k1k1k2k2[i-1], yaccM04_k1k1k2k2[i-1])
        right = gsl.interp2d_eval_no_boundary_check(interpM04_k1k1k2k2[i], &ktab[0], &ktab[0], &MT04_k1k1k2k2[i,0,0], k1, k2, xaccM04_k1k1k2k2[i], yaccM04_k1k1k2k2[i])
            
        return t*left + (1-t)*right


cpdef double M13_k1k1k2(double k1, double k2, double z):
    
    if z >= zmax :
        return gsl.interp2d_eval_no_boundary_check(interpM13_k1k1k2[znum-1], &ktab[0], &ktab[0], &MT13_k1k1k2[znum-1,0,0], k1, k2, xaccM13_k1k1k2[znum-1], yaccM13_k1k1k2[znum-1])

    else :
        i = search(ztabc, znum, z)
        t = (z - ztabc[i])/(ztabc[i-1] - ztabc[i])
    
        left = gsl.interp2d_eval_no_boundary_check(interpM13_k1k1k2[i-1], &ktab[0], &ktab[0], &MT13_k1k1k2[i-1,0,0], k1, k2, xaccM13_k1k1k2[i-1], yaccM13_k1k1k2[i-1])
        right = gsl.interp2d_eval_no_boundary_check(interpM13_k1k1k2[i], &ktab[0], &ktab[0], &MT13_k1k1k2[i,0,0], k1, k2, xaccM13_k1k1k2[i], yaccM13_k1k1k2[i])
            
        return t*left + (1-t)*right

    






### Trispectrum matter/galaxy in the halo model evaluated at points (k1, -k1, k2, -k2)


cpdef double trispectrum(double k1, double k2, double u12, double z) :
    
    #fonctions qui apparaîtront souvent
    
    cdef double M11k1 = M11_k1(k1, z)
    cdef double M11k2 = M11_k1(k2, z)
    
    cdef double M21k1 = M21_k1(k1, z)
    cdef double M21k2 = M21_k1(k2, z)
    
    cdef double M31k1 = M31_k1(k1, z)
    cdef double M31k2 = M31_k1(k2, z)
    
    cdef double M12k1k2 = M12_k1k2(k1, k2, z)
    
    cdef double M22k1k2 = M22_k1k2(k1, k2, z)
    
    
    cdef double n12 = ptc.norme2(k1, k2, u12)
    cdef double m12 = ptc.norme2(k1, k2, -u12)
    
    cdef double Pk1 = ptc.P(k1, z)
    cdef double Pk2 = ptc.P(k2, z)
    cdef double Pk1plusk2 = ptc.P(n12, z)
    cdef double Pk1moinsk2 = ptc.P(m12, z)
    
    cdef double u1_n12, u2_n12, u1_m12, u2_m12
    
    if n12 == 0 :
        u1_n12 = 0.
        u2_n12 = 0.
    else :
        u1_n12 = (k1*k1 + k1*k2*u12)/(k1 * n12)      #angle between k1 and k1+k2, etc
        u2_n12 = (k2*k2 + k1*k2*u12)/(k2 * n12)
        
    if m12 == 0 :
        u1_m12 = 0.
        u2_m12 = 0.
    else :
        u1_m12 = (k1*k1 - k1*k2*u12)/(k1 * m12)     
        u2_m12 = (-k2*k2 + k1*k2*u12)/(k2 * m12)

    
    
    cdef double T1h = M04_k1k1k2k2(k1, k2, z)
    
    
    
    cdef double T2h = (
    
          2 * M11k1 * M13_k1k1k2(k2, k1, z) * Pk1
        + 2 * M11k2 * M13_k1k1k2(k1, k2, z) * Pk2
        
        + M12k1k2 * M12k1k2 * (Pk1plusk2 + Pk1moinsk2)
    )
    
    
    
    cdef double T3h = (
    
          ( 2 * M11k1 * M12k1k2 * M11k2 ) * ptc.BPT(k1, k2, n12, u12, -u1_n12, -u2_n12, z)
        + ( 2 * M12k1k2 * M11k1 * M11k2 ) * ptc.BPT(k1, k2, m12, -u12, -u1_m12, u2_m12, z)
    
    
    
        + M11k1 * M11k1 * M22_k1k2(k2, k2, z) * Pk1*Pk1
        
        + 4 * M11k1 * M22k1k2 * M11k2 * Pk1 * Pk2
        
        + 2 * M21k1 * M12k1k2 * M11k2 * Pk2 * (Pk1plusk2 + Pk1moinsk2)
        + 2 * M11k1 * M12k1k2 * M21k2 * Pk1 * (Pk1plusk2 + Pk1moinsk2)
                        
        + M22_k1k2(k1, k1, z) * M11k2 * M11k2 * Pk2 * Pk2
    
    )
    
    cdef double T4h = (
        
          M11k1 * M11k1 * M11k2 * M11k2 * ptc.TPT(k1, k2, u12, z)
          
        + 2 * M31k1 * M11k1 * M11k2 * M11k2 * Pk1 * Pk2 * Pk2
        + 2 * M31k2 * M11k2 * M11k1 * M11k1 * Pk2 * Pk1 * Pk1
        
        + 2 * M21k1 * M11k1 * M11k2 * M11k2 * Pk2 * (ptc.BPT(k1, k2, n12, u12, -u1_n12, -u2_n12, z) + ptc.BPT(k1, k2, m12, -u12, -u1_m12, u2_m12, z))
        + 2 * M21k2 * M11k2 * M11k1 * M11k1 * Pk1 * (ptc.BPT(k1, k2, n12, u12, -u1_n12, -u2_n12, z) + ptc.BPT(k1, k2, m12, -u12, -u1_m12, u2_m12, z))

        + 2 * M21k1 * M11k2 * Pk2 * (M21k1 * M11k2 * Pk2 + M11k1 * M21k2 * Pk1) * (Pk1plusk2 + Pk1moinsk2)
        + 2 * M21k2 * M11k1 * Pk1 * (M21k2 * M11k1 * Pk1 + M11k2 * M21k1 * Pk2) * (Pk1plusk2 + Pk1moinsk2)
    
    )
 

    return T1h + T2h + T3h + T4h
    
   
    

###Powerspectrum in the halo model


cpdef double Pmm(double k, double z) :
    
    cdef double P1h = M02_k1k1(k, z)
    cdef double P2h = M11_k1(k, z)**2 * ptc.P(k, z)
    
    return P1h + P2h